#define DOUBLE(x) ((double)(x))

static __global__ void vecMul(cuComplex *a, cuComplex *b, int size) {
    float tmp;
    
    int i = threadIdx.x + blockIdx.x*size;
    
    tmp = a[i].x * b[i].x - a[i].y * b[i].y;
    a[i].y = a[i].x * b[i].y + a[i].y * b[i].x; 
    a[i].x = tmp;
}


static __global__ void vecPack(cufftReal *a, int asize, uint8_t *b, int bsize, int size) {
//    int i = threadIdx.x + blockIdx.x*bsize;
//    int i = bsize - threadIdx.x - 1 + (bsize - blockIdx.x - 1)*bsize;

	// Invalid memory access (possibly)
//    int i = (bsize - blockIdx.x)*bsize - threadIdx.x - 1;
    int i = size - threadIdx.x - 1 + (size - blockIdx.x - 1)*bsize;
    a[threadIdx.x + asize*blockIdx.x] = b[i];
}

static __global__ void vecPackBase(
    uint8_t *b, int bsize,
    cufftReal *a, int asize, 
    float *c, float *c2, int csize, int coffset) {
    
    float v;
    int i = threadIdx.x + (blockIdx.x+coffset)*csize + coffset;
    
    v = b[threadIdx.x + blockIdx.x*bsize];

    a[threadIdx.x + asize*blockIdx.x] = v;
    
    c[i] = v;
    c2[i] = v * v;
}


static __global__ void stat1(int *buf1, int *buf2, uint8_t *img, int image_pitch, int row_pitch, int size) {
    int i;
    int end = size * row_pitch;

    int side_idx =  blockIdx.x * blockDim.x + threadIdx.x;
    int img_idx = blockIdx.y * blockDim.y + threadIdx.y;

    int sum = 0;
    int sum2 = 0;

    uint8_t *vec = img + img_idx * image_pitch + side_idx;

    for (i = 0; i < end; i+=row_pitch) {
	int val = vec[i];
	sum += val;
	sum2 += val*val;
    }

    buf1[side_idx * CP_BLOCK + img_idx] = sum;
    buf2[side_idx * CP_BLOCK + img_idx] = sum2;
}

static __global__ void stat2(float *res1, float *res2, int *buf1, int *buf2, int size) {
    int i;
    int end = size * CP_BLOCK;
    int img_idx =  blockIdx.x * blockDim.x + threadIdx.x;

    int sum = 0;
    int sum2 = 0;

    int *vec1 = buf1 + img_idx;
    int *vec2 = buf2 + img_idx;

    for (i = 0; i < end; i+=CP_BLOCK) {
	sum += vec1[i];
	sum2 += vec2[i];
    }

    res1[img_idx] = sum;

    float cnt = size * size;
    float mean = ((float)sum) / cnt;
    
    res2[img_idx] = sqrtf(fmaxf(((float)sum2) / cnt - mean*mean,0));
}





/*
static __global__ void vecCompute(
    uint16_t *items, float *res,
    cufftReal *corr, float corr_scale, 
    float *lsum, float lsum_scale,
    float *denom, float denom_scale
) {
    int pos = items[threadIdx.x + blockIdx.x*BLOCK_SIZE_1D];//correct (row/column)?
    res[pos] = (corr[pos] * corr_scale - lsum[pos]*lsum_scale) / (denom[pos] * denom_scale);
}
*/

/*
static __global__ void vecCompute(
    float *res,
    cufftReal *corr, float corr_scale, 
    float *lsum, float lsum_scale,
    float *denom, float denom_scale,
    int size
) {
    int pos = threadIdx.x + blockIdx.x*size;

    if (denom[pos]) {
	res[pos] = (corr[pos] * corr_scale - lsum[pos]*lsum_scale) / (denom[pos] * denom_scale);
    }
}
*/

static __global__ void vecCompute(
    float *res,
    cufftReal *corr, float corr_scale, 
    float *lsum, float *lsum_scale_ptr, float lsum_mult,
    float *denom, float *denom_scale_ptr,
    int size
) {
    int pos = threadIdx.x + blockIdx.x*size;

    float lsum_scale = (*lsum_scale_ptr) * lsum_mult;
    float denom_scale = (*denom_scale_ptr);

    if (denom[pos]&&denom_scale) {
	res[pos] = (corr[pos] * corr_scale - lsum[pos]*lsum_scale) / (denom[pos] * denom_scale);
    }
}
